jitCholInv <- function ( M, Num=10, silent=FALSE ) {
  jitter <- 0
  jitter1 <- abs(mean(diag(M)))*1e-6
  eyeM <- diag( 1, nrow=length(M[,1]), ncol=length(M[1,]) )

  for ( i in 1:Num ) {

    ## clear the last error message
    try(stop(""),TRUE)
    ow <- options("warn")
    options(warn=2)

    Ch <- try( chol( M + jitter*eyeM ), silent=TRUE )

    options(ow)
    nPos <- grep("not positive definite",  geterrmessage())

    if ( length(nPos) != 0 ) {
      jitter1 <- jitter1*10
      jitter <- jitter1

      if (! silent) {
        warnmsg <- paste("Matrix is not positive definite, adding",
                         signif(jitter,digits=4), "jitter!")
        warning(warnmsg)
      }
    }
    else break
  }

  ow <- options("warn")
  options(warn=2)
  invM <- try (chol2inv(Ch), silent=TRUE)
  options(ow)

  if ( class(invM) == "try-error" ) {
    return (list(invM=NaN))
  }
  else {
    if ( jitter == 0 ) {
      ans <- list(invM=invM, jitter=jitter, chol=Ch)
    }
    else ans <- list(invM=invM, jitM=M+jitter*eyeM , jitter=jitter, chol=Ch)

    return (ans)
  }
}


gpPlot2 <- function (model, Xstar, mu, S, simpose = NULL, xlim = NULL, ylim = NULL, 
    xlab = "", ylab = "", col = "blue", title = "") 
{
    if (missing(model) || missing(Xstar)) {
        stop("Missing GP model or points of prediction Xstar.")
    }
    else {
        if (missing(mu) || missing(S)) {
            meanVar = gpPosteriorMeanVar(model, Xstar, varsigma.return = TRUE)
            mu = meanVar$mu
            S = meanVar$varsigma
        }
    }
    f = c(mu + 2 * sqrt(abs(S)), rev(mu - 2 * sqrt(abs(S))))
    if (is.null(xlim)) 
        xlim = range(rbind(model$X, Xstar))
    if (is.null(ylim)) 
        ylim = range(f)
    plot(0, type = "n", xlim = xlim, ylim = ylim, xlab = xlab, 
        ylab = ylab, main = title, new = TRUE)
    if (col == "blue") 
        shade = rgb(0, 0, 1, alpha = 0.1)
    else if (col == "red") 
        shade = rgb(255, 0, 0, alpha = 0.1)
    else shade = "gray"
    polygon(c(Xstar, rev(Xstar)), f, col = shade, border = shade)
    lines(Xstar, mu, col = col, lwd = 2)
    if (!is.null(simpose)) {
        y = mu[simpose] + rnorm(6, 0, exp(model$params$xmin[3]/2))
        points(simpose, y, pch = 4, cex = 1.5, lwd = 3, col = col)
    }
}

gpPersp <- function (model, Xstar, mu, S, simpose = NULL, xlim = NULL, ylim = NULL, phi = 45, theta = 20,
    xlab = "", ylab = "", col = "blue", title = "") 
{
    if (missing(model) || missing(Xstar)) {
        stop("Missing GP model or points of prediction Xstar.")
    }
    else {
        if (missing(mu) || missing(S)) {
            meanVar = gpPosteriorMeanVar(model, Xstar, varsigma.return = TRUE)
            mu = meanVar$mu
            S = meanVar$varsigma
        }
    }
    f = c(mu + 2 * sqrt(abs(S)), rev(mu - 2 * sqrt(abs(S))))
    if (is.null(xlim)) 
        xlim = range(rbind(model$X, Xstar))
    if (is.null(ylim)) 
        ylim = range(f)

    yy <- seq(ylim[1], ylim[2], length.out=length(Xstar))
    zz <- c()
    for ( i in 1:length(Xstar)) {
      zz <- rbind(zz, dnorm(yy, mu[i], sqrt(S[i])))
    }
    nrz <- nrow(zz)
    ncz <- ncol(zz)
    # Create a function interpolating colors in the range of specified colors
    #jet.colors <- colorRampPalette( c("blue", "red") )
    heat.colorsX <- colorRampPalette( c("lightblue","yellow", "red") )
    # Generate the desired number of colors from this palette
    nbcol <- 100
    color <- heat.colorsX(nbcol)
    # Compute the z-value at the facet centres
    zfacet <- zz[-1, -1] + zz[-1, -ncz] + zz[-nrz, -1] + zz[-nrz, -ncz]
    # Recode facet z-values into color indices
    facetcol <- cut(zfacet, nbcol)
    persp(Xstar, yy, zz, col = color[facetcol], phi = phi, theta = theta, expand=.5, xlab="x*", ylab="f(x*)", zlab="", main=title, xlim=xlim, ylim=ylim)

}

plotSampleGP <- function(nsamples, xtest, kernfunc, tau=1, l=1, mu, xlim = NULL, ylim = NULL,
    xlab = "", ylab = "", col = "blue", title = "") 
{
  if (missing(mu)) {
    mu=rep(0, length(xtest))
  }
  sigma <- kernfunc(xtest,xtest, tau, l)
  S <- diag(sigma)

  values <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  for (i in 1:nsamples) {
    # Each column represents a sample from a multivariate normal distribution
    # with zero mean and covariance sigma
    values[,i] <- mvrnorm(1, rep(0, length(xtest)), sigma)
  }

  matplot(xtest, values, type='l', xlab="x*", ylab="f(x*)", main=title, xlim=xlim, ylim=ylim)
}

#gaussProcPersp <-  function(xtest, kernfunc, tau=1, l=1, mu, xlim = NULL, ylim = NULL, phi = 45, theta = 20, 
gaussProcPersp <-  function(xtest, sigma, mu, xlim = NULL, ylim = NULL, phi = 45, theta = 20, 
    xlab = "", ylab = "", col = "blue", title = "") 
{
  if (missing(mu)) {
    mu=rep(0, length(xtest))
  }
  #sigma <- kernfunc(xtest,xtest, tau, l)
  S <- diag(sigma)
  f = c(mu + 2 * sqrt(abs(S)), rev(mu - 2 * sqrt(abs(S))))
  if (is.null(xlim)) 
      xlim = range(xtest)
  if (is.null(ylim)) 
      ylim = range(f)
  
  yy <- seq(ylim[1], ylim[2], length.out=length(xtest))
  zz <- c()
  for ( i in 1:length(xtest)) {
    zz <- rbind(zz, dnorm(yy, mu[i], sqrt(S[i])))
  }
  nrz <- nrow(zz)
  ncz <- ncol(zz)
  # Create a function interpolating colors in the range of specified colors
  #jet.colors <- colorRampPalette( c("blue", "red") )
  heat.colorsX <- colorRampPalette( c("lightblue","yellow", "red") )
  # Generate the desired number of colors from this palette
  nbcol <- 100
  color <- heat.colorsX(nbcol)
  # Compute the z-value at the facet centres
  zfacet <- zz[-1, -1] + zz[-1, -ncz] + zz[-nrz, -1] + zz[-nrz, -ncz]
  # Recode facet z-values into color indices
  facetcol <- cut(zfacet, nbcol)
  persp(xtest, yy, zz, col = color[facetcol], phi = phi, theta = theta, expand=.5, xlab="x*", ylab="f(x*)", zlab="", main=title, xlim=xlim, ylim=ylim)

}
