library(gptk)
library(fMultivar)
source("kernels.R")
source("gpPlotFuns.R")
setEPS()


#### SLIDE 1 ####
if (FALSE) {
  x <- seq(-4,4, length.out=50)
  postscript("Univariate.eps", height=4, width=4)
    plot(x, dnorm(x), type='l', xlab='x', ylab='', main="Univariate Normal", cex.main=1.6)
  dev.off()
  
  y <- x
  z <- outer(x, y, dnorm2d, rho=0)
  
  heat.colorsX <- colorRampPalette( c("lightblue","yellow", "red") )
  nbcol <- 100
  color <- heat.colorsX(nbcol)
  # Compute the z-value at the facet centres
  nrz <- nrow(z)
  ncz <- ncol(z)
  zfacet <- z[-1, -1] + z[-1, -ncz] + z[-nrz, -1] + z[-nrz, -ncz]
  facetcol <- cut(zfacet, nbcol)
  
  postscript("Bivariate.eps", height=4, width=4)
    persp(x, y, z, theta = 30, phi = 30, expand = 0.5, col = color[facetcol], xlab='x', ylab='y', zlab='', main="Bivariate Normal")
  dev.off()
  
  
  options = gpOptions()
  options$kern$comp = list("rbf", "white")
  x = matrix(seq(-1, 1, length = 6), ncol = 1)
  xtest = matrix(seq(-1.5, 1.5, length = 50), ncol = 1)
  trueKern = kernCreate(x, list(type = "cmpnd", comp = list("rbf", 
        "white")))
  kern = trueKern
  K = kernCompute(trueKern, x)
  y = t(gaussSamp(matrix(0, 6, 1), K, 1))
  y = scale(y, scale = FALSE)
  model = gpCreate(dim(x)[2], dim(y)[2], x, y, options)
  
  postscript("GaussianProcess.eps", height=4, width=4)
    gpPersp(model, xtest, title="Gaussian Process")
  dev.off()
}

#### SLIDE 2 ####
if (FALSE) {
  options = gpOptions()
  options$kern = "white"
  x = matrix(seq(0, 2, length = 6), ncol = 1)
  xtest = matrix(seq(0, 3, length = 50), ncol = 1)
  trueKern = kernCreate(x, list(type = "white"))
  kern = trueKern
  K = kernCompute(trueKern, x)
  y = t(gaussSamp(matrix(0, 6, 1), K, 1))
  y = scale(y, scale = FALSE)
  model = gpCreate(dim(x)[2], dim(y)[2], x, y, options)
  
  postscript("WhiteNoise_persp.eps", height=4, width=4)
    gpPersp(model, xtest, phi=60, title="White Noise")
  dev.off()
  #system("epspdf WhiteNoise_persp.eps")

  postscript("BrownianMotion_persp.eps", height=4, width=4)
    gaussProcPersp(xtest, calcSigmaBrown(xtest, xtest, tau=.5), phi=20, title="Brownian Motion")
  dev.off()
  #system("epspdf BrownianMotion_persp.eps")

  postscript("FractionalBrownianMotion_persp.eps", height=4, width=4)
    gaussProcPersp(xtest=xtest, sigma=calcSigmaPink(xtest, xtest, tau=.5, l=.25), phi=20, title="Fractional Brownian Motion")
  dev.off()
  #system("epspdf FractionalBrownianMotion_persp.eps")


  xtest = matrix(seq(0, 3, length = 500), ncol = 1)

  postscript("WhiteNoise_samps.eps", height=4, width=4)
    plotSampleGP(4, xtest, calcSigmaWhite, tau=1)
  dev.off()
  #system("epspdf WhiteNoise_samps.eps")

  postscript("BrownianMotion_samps.eps", height=4, width=4)
    plotSampleGP(4, xtest, calcSigmaBrown, tau=.5)
  dev.off()
  #system("epspdf BrownianMotion_samps.eps")

  postscript("FractionalBrownianMotion_samps.eps", height=4, width=4)
    plotSampleGP(4, xtest, calcSigmaPink, tau=1, l=.25)
  dev.off()
  #system("epspdf FractionalBrownianMotion_samps.eps")
}


### SLIDE X:  BAYESIAN LINEAR REGRESSION
if(TRUE) {
  x = c(.5, 1.5, 2)
  y = c(-2, 0.5, 1)
  xtest = matrix(seq(0, 3, length = 100), ncol = 1)

  sigma.n <- .1

  k.xx <- calcSigmaLinear(x, x)
  k.xtx <- calcSigmaLinear(xtest, x)
  k.xtxt <- calcSigmaLinear(xtest, xtest)
  k.xxt <- calcSigmaLinear(x, xtest)
  
  mu.post <- k.xtx %*% solve(k.xx + sigma.n^2*diag(1, ncol(k.xx))) %*% y
  cov.post <- k.xtxt - k.xtx %*% solve(k.xx + sigma.n^2*diag(1, ncol(k.xx))) %*%  k.xxt

}


#### SLIDE  X:  INTERPOLATION ####
if(FALSE) {
  x = c(.5, 1.5, 2, 1.75)
  y = c(-2, 0, -1, .5)
  xtest = matrix(seq(0, 3, length = 100), ncol = 1)

# By hand (following Keirstead based on GPML eqn2.19)
  k.xx <- calcSigmaRBF(x,x, tau=1, l=3)
  k.xxt <- calcSigmaRBF(x,xtest, tau=1, l=3)
  k.xtx <- calcSigmaRBF(xtest,x, tau=1, l=3)
  k.xtxt <- calcSigmaRBF(xtest,xtest, tau=1, l=3)

  mu.post <- k.xtx%*%solve(k.xx)%*%y
  cov.post <- k.xtxt - k.xtx%*%solve(k.xx)%*%k.xxt

  nsamples <- 50
  prior <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  values <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  for (i in 1:nsamples) {
    prior[,i] <- mvrnorm(1, rep(0, length(xtest)), k.xtxt)
    values[,i] <- mvrnorm(1, mu.post, cov.post)
  }

  postscript("Interpolation.eps", height=4, width=8)
    par(mfrow=c(1,2))
    matplot(xtest, prior, type='l', main="Gaussian Process Prior", xlab="x*", ylab="f(x*)")
    matplot(xtest, values, type='l', main="Gaussian Process Interpolation", xlab="x*", ylab="f(x*)")
    points(x, y, pch="+", cex=2.5, col='green')
  dev.off()

# Using gptk (From gptk::demInterterpoloation)
  yTrain = as.matrix(y)
  xTrain = as.matrix(x)
  kern = kernCreate(xtest, "rbf")
  kern$inverseWidth = 1
  Kx = kernCompute(kern, xtest, xTrain)
  Ktrain = kernCompute(kern, xTrain, xTrain)
  invKtrain = jitCholInv(Ktrain, silent = TRUE)$invM
  yPred = Kx %*% invKtrain %*% yTrain
  yVar = kernDiagCompute(kern, xtest) - rowSums(Kx %*% 
      invKtrain * Kx)

  model = gpCreate(dim(xTrain)[2], dim(yTrain)[2], xTrain, 
      yTrain, options)
  gpPlot(model, xtest, yPred, yVar, ylim = c(-4, 1.5), col = "black")

}


### SLIDE X:  REGRESSION ###
if(FALSE) {
  x = c(.5, 1.5, 2)
  y = c(-2, 0, -1)
  xtest = matrix(seq(0, 3, length = 100), ncol = 1)
  
# By hand (following Keirstead based on GPML)
  sigma.n <- 0.1
  k.xx <- calcSigmaRBF(x,x)
  k.xxt <- calcSigmaRBF(x,xtest)
  k.xtx <- calcSigmaRBF(xtest,x)
  k.xtxt <- calcSigmaRBF(xtest,xtest)

  #mu.post <- k.xtx%*%solve(k.xx)%*%y
  mu.post <- k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%y
  #cov.post <- k.xtxt - k.xtx%*%solve(k.xx)%*%k.xxt
  cov.post <- k.xtxt - k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%k.xxt

  nsamples <- 50
  prior <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  values <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  for (i in 1:nsamples) {
    prior[,i] <- mvrnorm(1, rep(0, length(xtest)), k.xtxt)
    values[,i] <- mvrnorm(1, mu.post, cov.post)
  }

  postscript("Regression.eps", height=4, width=8)
    par(mfrow=c(1,2))
    matplot(xtest, prior, type='l', main="Gaussian Process Prior", xlab="x*", ylab="f(x*)")
    matplot(xtest, values, type='l', main="Gaussian Process Regression", xlab="x*", ylab="f(x*)")
    points(x, y, pch="+", cex=2.5, col='green')
  dev.off()
  gaussProcPersp(xtest=xtest, sigma=cov.post, mu=round(mu.post,4), phi=20, title="Gaussian Process Regression")

# Using gptk  (From gptk::demRegression)
  yTrain = as.matrix(y)
  xTrain = as.matrix(x)
  kern = kernCreate(xtest, "rbf")
  kern$inverseWidth = 1
  Kx = kernCompute(kern, xtest, xTrain)
  Ktrain = kernCompute(kern, xTrain, xTrain)
  invKtrain = jitCholInv(Ktrain + diag(dim(Ktrain)[1]) * 
                sigma.n^2, silent = TRUE)$invM
  yPred = Kx %*% invKtrain %*% yTrain
  yVar = kernDiagCompute(kern, xtest) - rowSums(Kx %*% 
      invKtrain * Kx)

  model = gpCreate(dim(xTrain)[2], dim(yTrain)[2], xTrain, 
      yTrain, options)
  gpPlot(model, xtest, yPred, yVar, ylim = c(-4, 1.5), col = "black")
  gpPersp(model, xtest, yPred, yVar, ylim = c(-4, 1.5), col = "black")

}


## BAYESIAN LINEAR REGRESSION
if(FALSE) {
  x = c(.5, 1.1, 1.5, 2)
  y = c(0, .5, .6, 1)
  xtest = matrix(seq(0, 3, length = 100), ncol = 1)
  
# By hand (following Keirstead based on GPML)
  sigma.n <- 0.1
  k.xx <- calcSigmaLinear(x,x)
  k.xxt <- calcSigmaLinear(x,xtest)
  k.xtx <- calcSigmaLinear(xtest,x)
  k.xtxt <- calcSigmaLinear(xtest,xtest)

  #mu.post <- k.xtx%*%solve(k.xx)%*%y
  mu.post <- k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%y
  #cov.post <- k.xtxt - k.xtx%*%solve(k.xx)%*%k.xxt
  cov.post <- k.xtxt - k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%k.xxt
  S <- sqrt(diag(cov.post))

  nsamples <- 50
  prior <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  values <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  for (i in 1:nsamples) {
    prior[,i] <- mvrnorm(1, rep(0, length(xtest)), k.xtxt)
    values[,i] <- mvrnorm(1, mu.post, cov.post)
  }

  postscript("LinearRegression.eps", height=4, width=12)
    par(mfrow=c(1,3))
    matplot(xtest, prior, type='l', main="Prior Samples", xlab="x*", ylab="f(x*)")

    matplot(xtest, values, type='l', main="Posterior Samples", xlab="x*", ylab="f(x*)")
    points(x, y, pch="+", cex=2.5, col='green')

    plot(xtest, mu.post, type='n',  main="Point-wise Posterior Uncertainty", xlab="x*", ylab="f(x*)")
    polygon(c(xtest,rev(xtest)), c(mu.post+2*S, rev(mu.post-2*S)), col='grey70', border=NA)
    lines(xtest, mu.post)
  dev.off()
  gaussProcPersp(xtest=xtest, sigma=cov.post, mu=round(mu.post,4), phi=20, title="Gaussian Process Regression")
}
## RBF REGRESSION
if(FALSE) {
  x = c(.5, 1.5, 2)
  y = c(-2, 0, -1)
  xtest = matrix(seq(0, 3, length = 100), ncol = 1)
  
# By hand (following Keirstead based on GPML)
  sigma.n <- 0.1
  k.xx <- calcSigmaRBF(x,x)
  k.xxt <- calcSigmaRBF(x,xtest)
  k.xtx <- calcSigmaRBF(xtest,x)
  k.xtxt <- calcSigmaRBF(xtest,xtest)

  mu.post <- k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%y
  cov.post <- k.xtxt - k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%k.xxt
  S <- sqrt(diag(cov.post))

  nsamples <- 50
  prior <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  values <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  for (i in 1:nsamples) {
    prior[,i] <- mvrnorm(1, rep(0, length(xtest)), k.xtxt)
    values[,i] <- mvrnorm(1, mu.post, cov.post)
  }

  postscript("RBFRegression.eps", height=4, width=12)
    par(mfrow=c(1,3))
    matplot(xtest, prior, type='l', main="Prior Samples", xlab="x*", ylab="f(x*)")

    matplot(xtest, values, type='l', main="Posterior Samples", xlab="x*", ylab="f(x*)")
    points(x, y, pch="+", cex=2.5, col='green')

    
    #plot(xtest, mu.post, type='n',  main="Point-wise Posterior Uncertainty", xlab="x*", ylab="f(x*)")
    plot(c(xtest,rev(xtest)), c(mu.post+2*S, rev(mu.post-2*S)), type='n', main="Point-wise Posterior Uncertainty", xlab="x*", ylab="f(x*)")
    polygon(c(xtest,rev(xtest)), c(mu.post+2*S, rev(mu.post-2*S)), col='grey70', border=NA)
    lines(xtest, mu.post)
  dev.off()
  gaussProcPersp(xtest=xtest, sigma=cov.post, mu=round(mu.post,4), phi=20, title="Gaussian Process Regression")
}



## MATERN REGRESSION
if(FALSE) {
  x = c(.5, 1.5, 2)
  y = c(-2, 0, -1)
  xtest = matrix(seq(0, 3, length = 100), ncol = 1)
  
# By hand (following Keirstead based on GPML)
  sigma.n <- 0.1
  k.xx <- calcSigmaMatern(x,x, nu=3/2)
  k.xxt <- calcSigmaMatern(x,xtest, nu=3/2)
  k.xtx <- calcSigmaMatern(xtest,x, nu=3/2)
  k.xtxt <- calcSigmaMatern(xtest,xtest, nu=3/2)

  mu.post <- k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%y
  cov.post <- k.xtxt - k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%k.xxt
  S <- sqrt(diag(cov.post))

  nsamples <- 50
  prior <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  values <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  for (i in 1:nsamples) {
    prior[,i] <- mvrnorm(1, rep(0, length(xtest)), k.xtxt)
    values[,i] <- mvrnorm(1, mu.post, cov.post)
  }

  postscript("MaternRegression.eps", height=4, width=12)
    par(mfrow=c(1,3))
    matplot(xtest, prior, type='l', main="Prior Samples", xlab="x*", ylab="f(x*)")

    matplot(xtest, values, type='l', main="Posterior Samples", xlab="x*", ylab="f(x*)")
    points(x, y, pch="+", cex=2.5, col='green')

    
    #plot(xtest, mu.post, type='n',  main="Point-wise Posterior Uncertainty", xlab="x*", ylab="f(x*)")
    plot(c(xtest,rev(xtest)), c(mu.post+2*S, rev(mu.post-2*S)), type='n', main="Point-wise Posterior Uncertainty", xlab="x*", ylab="f(x*)")
    polygon(c(xtest,rev(xtest)), c(mu.post+2*S, rev(mu.post-2*S)), col='grey70', border=NA)
    lines(xtest, mu.post)
  dev.off()
  gaussProcPersp(xtest=xtest, sigma=cov.post, mu=round(mu.post,4), phi=20, title="Gaussian Process Regression")
}


## RQ REGRESSION
if(FALSE) {
  x = c(.5, 1.5, 2)
  y = c(-2, 0, -1)
  xtest = matrix(seq(0, 3, length = 100), ncol = 1)
  
# By hand (following Keirstead based on GPML)
  sigma.n <- 0.1
  k.xx <- calcSigmaRQ(x,x)
  k.xxt <- calcSigmaRQ(x,xtest)
  k.xtx <- calcSigmaRQ(xtest,x)
  k.xtxt <- calcSigmaRQ(xtest,xtest)

  mu.post <- k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%y
  cov.post <- k.xtxt - k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%k.xxt
  S <- sqrt(diag(cov.post))

  nsamples <- 50
  prior <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  values <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  for (i in 1:nsamples) {
    prior[,i] <- mvrnorm(1, rep(0, length(xtest)), k.xtxt)
    values[,i] <- mvrnorm(1, mu.post, cov.post)
  }

  postscript("RQRegression.eps", height=4, width=12)
    par(mfrow=c(1,3))
    matplot(xtest, prior, type='l', main="Prior Samples", xlab="x*", ylab="f(x*)")

    matplot(xtest, values, type='l', main="Posterior Samples", xlab="x*", ylab="f(x*)")
    points(x, y, pch="+", cex=2.5, col='green')

    
    #plot(xtest, mu.post, type='n',  main="Point-wise Posterior Uncertainty", xlab="x*", ylab="f(x*)")
    plot(c(xtest,rev(xtest)), c(mu.post+2*S, rev(mu.post-2*S)), type='n', main="Point-wise Posterior Uncertainty", xlab="x*", ylab="f(x*)")
    polygon(c(xtest,rev(xtest)), c(mu.post+2*S, rev(mu.post-2*S)), col='grey70', border=NA)
    lines(xtest, mu.post)
  dev.off()
  gaussProcPersp(xtest=xtest, sigma=cov.post, mu=round(mu.post,4), phi=20, title="Gaussian Process Regression")
}

## PERIODIC REGRESSION
if(FALSE) {
  x = c(.5, 1.5, 2)
  y = c(-2, 0, -1)
  xtest = matrix(seq(0, 3, length = 100), ncol = 1)
  
# By hand (following Keirstead based on GPML)
  sigma.n <- 0.1
  k.xx <- calcSigmaPeriodic(x,x)
  k.xxt <- calcSigmaPeriodic(x,xtest)
  k.xtx <- calcSigmaPeriodic(xtest,x)
  k.xtxt <- calcSigmaPeriodic(xtest,xtest)

  mu.post <- k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%y
  cov.post <- k.xtxt - k.xtx%*%solve(k.xx + sigma.n^2*diag(1, ncol(k.xx)))%*%k.xxt
  S <- sqrt(diag(cov.post))

  nsamples <- 10
  prior <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  values <- matrix(rep(0,length(xtest)*nsamples), ncol=nsamples)
  for (i in 1:nsamples) {
    prior[,i] <- mvrnorm(1, rep(0, length(xtest)), k.xtxt)
    values[,i] <- mvrnorm(1, mu.post, cov.post)
  }

  postscript("PeriodicRegression.eps", height=4, width=12)
    par(mfrow=c(1,3))
    matplot(xtest, prior, type='l', main="Prior Samples", xlab="x*", ylab="f(x*)")

    matplot(xtest, values, type='l', main="Posterior Samples", xlab="x*", ylab="f(x*)")
    points(x, y, pch="+", cex=2.5, col='green')

    
    #plot(xtest, mu.post, type='n',  main="Point-wise Posterior Uncertainty", xlab="x*", ylab="f(x*)")
    plot(c(xtest,rev(xtest)), c(mu.post+2*S, rev(mu.post-2*S)), type='n', main="Point-wise Posterior Uncertainty", xlab="x*", ylab="f(x*)")
    polygon(c(xtest,rev(xtest)), c(mu.post+2*S, rev(mu.post-2*S)), col='grey70', border=NA)
    lines(xtest, mu.post)
  dev.off()
  gaussProcPersp(xtest=xtest, sigma=cov.post, mu=round(mu.post,4), phi=20, title="Gaussian Process Regression")
}
